/*********************************************************************
                       anselm.su [2013-05-30]

CREATE or replace FUNCTION bprintf(text, bytea) RETURNS text
     AS '/usr/local/lib/bytea', 'bprintf'
     LANGUAGE C STRICT;


CREATE or replace FUNCTION bscanf(text, text) RETURNS bytea
     AS '/usr/local/lib/bytea', 'bscanf'
     LANGUAGE C STRICT;
         
psql -Upgsql db1 -c"select bprintf( '%f', bscanf('11.33', '%f') )"

*********************************************************************/
#include <postgres.h>
#include <fmgr.h>

#ifdef PG_MODULE_MAGIC
PG_MODULE_MAGIC;
#endif


#define BLOCK_SPRINTF \
  if(b+sizeof(Type)>b0+size_data) \
    result = -1; \
  else { \
    result = sprintf(ret, left, *(Type*)b ); \
    b += sizeof(Type); \
  }

#define BLOCK_SSCANF \
  memcpy(b, tmp, sizeof(Type)); \
  b += sizeof(Type);  
  
#define MAX_STR_SIZE 255  

int bnprintf_temp(const void *data, const size_t size_data, const char *format, char *ret, const size_t size_ret)
{
  //const int size_ret = sizeof(ret); //sizeof не распознаёт реальный размер!!!
  char *left  = (char*) malloc(size_ret);
  char *right = (char*) malloc(size_ret);  
  strcpy(ret, format);  
  const char *b = (const char*) data, *b0=b;

  char *pos;
  while( pos=strchr(ret, '%') ) {
    int i = pos-ret;
    if(i>=size_ret-1) { // если '%' стоит в самом конце строки и даже дальше
      memset(ret, 0, size_ret); //strcpy(ret, "error");
      break;
    }
    
    char *pos2 = strpbrk(&ret[i+1], "eEfFgGaAdiouxXs");
    if(pos2==NULL) { // это может случиться в случае неправильного формата
      memset(ret, 0, size_ret);//strcpy(ret, "error");//
      break;
    }
    int j = pos2-ret;

    memset(left,  0, size_ret);
    memset(right, 0, size_ret);
    strncpy(left, ret, j+1);
    strcpy(right, &ret[j+1]);

    int result;
  
    if( strchr("eEfFgGaA",ret[j]) ) {
      if(ret[j-1]=='l') {
        typedef double Type;
        BLOCK_SPRINTF;
      }
      else {
        typedef float Type;
        BLOCK_SPRINTF;
      }
    }
    else if( strchr("di",ret[j]) ) {
      if(ret[j-1]=='l') {
        typedef long Type;
        BLOCK_SPRINTF;
      }
      else {
        typedef int Type;
        BLOCK_SPRINTF;
      }
    }
    else if( strchr("ouxX",ret[j]) ) {
      if(ret[j-1]=='l') {
        typedef unsigned long Type;
        BLOCK_SPRINTF;
      }
      else {
        typedef unsigned int Type;
        BLOCK_SPRINTF;
      }
    }
    else if( strchr("s",ret[j]) ) {
      if(ret[j-1]=='l') {
        int len = pg_wchar2mb(b, ret) * sizeof(wchar_t);
        if(b+len>b0+size_data) {
          //elog(NOTICE, "error: len=%d\n", len);
          result = -1;
        }
        else {
          left[j-1] = 's'; left[j] = 0;
          char *tmp = strdup(ret);
          result = sprintf(ret, left, tmp);
          b += len+sizeof(wchar_t);
          free(tmp);
        }
      }
      else {
        if(b+strlen(b)>b0+size_data)
          result = -1;
        else {
          result = sprintf(ret, left, b);
          b += strlen(b)+1;
        }
      }
    }
    else {
      result = -1; // всё остальное форматирование пока запретим
    }
    
    if(result==-1) {
      memset(ret, 0, size_ret);//strcpy(ret, "error");//
      break;
    }
    
    strcat(ret, right);
  }
  
  free(left);
  free(right);
  
  return strlen(ret);
}


PG_FUNCTION_INFO_V1(bprintf);

Datum
bprintf(PG_FUNCTION_ARGS)
{
  text  *arg1 = PG_GETARG_TEXT_P(0);   // формат - строка
  bytea *arg2 = PG_GETARG_BYTEA_P(1);  // данные - последовательность байт
  
  char *format = VARDATA_ANY(arg1);
  int size_format = VARSIZE_ANY_EXHDR(arg1);
  format[size_format] = 0;             // отсекаем мусор
  
  int32 new_text_size = VARSIZE(arg1) + VARSIZE(arg2)*sizeof(wchar_t); //TODO как правильно предположить размер?
  text *new_text = (text *) palloc(new_text_size);
  
  new_text_size = VARHDRSZ + bnprintf_temp(VARDATA(arg2), VARSIZE(arg2)-VARHDRSZ, VARDATA_ANY(arg1), VARDATA_ANY(new_text), new_text_size);
  SET_VARSIZE(new_text, new_text_size);
    
  PG_RETURN_TEXT_P(new_text);  // отформатированные данные
}

/*char* insert_pos(char *s, const char c, const char* pos)
{
  const size_t n = strlen(s);
  size_t i = pos-s;
  i = i>n || i<0 ? n : i;
  
  char *a = strdup(s);
  memset(s, 0, strlen(s));
  strncpy(s, a, i);
  char x[] = {c, 0};
  strcat(s, x);
  strcat(s, a+i);
  free(a);
  return s;
}*/

// Вставка символа в строку в позицию pos
char* insert_pos(char *s, const char c, const char * const pos)
{
  const size_t n = strlen(s);
  char * const end = s+n;
  
  char *p;
  
  if(pos<s)
    p = s;
  else if(pos>end)
    p = end;
  else
    p = s + (pos-s);
  
  char a = *p;
  *p = c;
  while(p++<end) {
    char b = *p;
    *p = a;
    a = b;
  }
  *p = 0;
  
  return s;
}


//Распознаёт i-тый неэкранированный символ '%'
const char* sscanf1(const char *s, const int num, const char *format, void *x)
{
  const char *ret = NULL;
  const size_t size_format = strlen(format);
  
  const size_t size_f = size_format*2+1; // с учётом добавления '*'
  char *f = (char*)malloc(size_f), * const f0 = f;
  memset(f, 0, size_f);
  strcpy(f, format);
  
  char *pos;  
  int i = 0;
  while( (pos=strchr(f, '%')) && pos-f0<size_f ) {
    
    f = pos+1;
    if( *f=='*' || *f=='%' ) // пропускаем экранирующие символы
      f++;
    else {
      //i++; // счёт ведём с единицы
      if(i!=num) {
        insert_pos(f0, '*', f);         
      }
      else {
        int j = f-1-f0-i;
        ret = format+j;
      }
      i++;   // счёт ведём с нуля
    }//else
    
  }//while
  
  sscanf(s, f0, x);
  //elog(NOTICE, "sscanf1(%s,%s,%ls)\n", s, f0, x);
  free(f0);
  return ret; // указатель на '%'
}

int bscanf_temp(const char *data, const char *format, char *ret, const size_t size_ret)
{
  char *b = (char*) ret, * const b0=b;
  
  //const size_t size_ret = sizeof(ret); //sizeof не распознаёт реальный размер!!!
  memset(ret, 0, size_ret);              // очистим ret от мусора
  //elog(NOTICE, "size_ret=%d\n", size_ret);
  
  const char *pos;
  int i = 0;
  char tmp[MAX_STR_SIZE] = {0};
  while( pos=sscanf1(data, i++, format, tmp) ) {
    
    char *pos2 = strpbrk(pos, "eEfFgGaAdiouxXs");
    if(pos2==NULL) { // это может случиться в случае неправильного формата
      memset(b=b0, 0, size_ret);//strcpy(f, "error");//
      break;
    }
    
    if( strchr("eEfFgGaA", *pos2) ) {
      if( *(pos2-1)=='l' ) {
        typedef double Type;
        BLOCK_SSCANF
      }
      else {
        typedef float Type;
        BLOCK_SSCANF
      }
    }
    else if( strchr("di", *pos2) ) {
      if( *(pos2-1)=='l' ) {
        typedef long Type;
        BLOCK_SSCANF
      }
      else {
        typedef int Type;
        BLOCK_SSCANF
      }
    }
    else if( strchr("ouxX", *pos2) ) {
      if( *(pos2-1)=='l' ) {
        typedef unsigned long Type;
        BLOCK_SSCANF
      }
      else {
        typedef unsigned int Type;
        BLOCK_SSCANF
      }
    }
    else if( strchr("s", *pos2) ) {
      if( *(pos2-1)=='l' ) {
      	int len = wcslen((wchar_t*)tmp)*sizeof(wchar_t);
      	//int len = pg_wchar2mb(b, ret);
        memcpy(b, tmp, len);
        //elog(NOTICE, "tmp=%ls, b=%ls, len=%d\n", tmp, b, len);
			  b += len+sizeof(wchar_t);
      }
      else {
      	int len = strlen(tmp);
        memcpy(b, tmp, len);
			  b += len+1;
      }
    }
    else {
      memset(b=b0, 0, size_ret); // всё остальное форматирование пока запретим
    }
    
    memset(tmp, 0, MAX_STR_SIZE); // очистим временный буфер
  }//while
  
  //elog(NOTICE, "ret=%ls:%d=[%X,%X,%X,%X]\n", b0, b-b0, b0[0]&0xff,b0[1]&0xff,b0[2]&0xff,b0[3]&0xff);
  return b-b0;
}

PG_FUNCTION_INFO_V1(bscanf);

Datum
bscanf(PG_FUNCTION_ARGS)
{
  text *arg1 = PG_GETARG_TEXT_P(0);   // данные - строка
  text *arg2 = PG_GETARG_TEXT_P(1);   // формат - строка

  char *data = VARDATA_ANY(arg1);
  int size_data = VARSIZE_ANY_EXHDR(arg1);
  data[size_data] = 0;                // отсекаем мусор
  
  char *format = VARDATA_ANY(arg2);
  int size_format = VARSIZE_ANY_EXHDR(arg2);
  format[size_format] = 0;            // отсекаем мусор
  
  int32 new_bytea_size = VARSIZE(arg1) * sizeof(wchar_t); // размер данных в текстовом виде точно больше размера в двоичном виде
  //elog(NOTICE, "new_bytea_size=%d", new_bytea_size);
  bytea *new_bytea = (bytea *) palloc(new_bytea_size);

  new_bytea_size = VARHDRSZ + bscanf_temp(data, format, VARDATA(new_bytea), new_bytea_size);
  SET_VARSIZE(new_bytea, new_bytea_size);
  	
  PG_RETURN_BYTEA_P(new_bytea);       // последовательность байт      
}
